import re
from collections import namedtuple
from random import randint, uniform, choice

from acrylic.color_names import color_names


Rgb = namedtuple('Rgb', 'r g b')
Hsl = namedtuple('Hsl', 'h s l')
Hsv = namedtuple('Hsv', 'h s v')
Ryb = namedtuple('Ryb', 'r y b')


Schema = namedtuple(
    'Schema',
    'format names length rng input_type output_type validation_type'
)


RANDOM = -1
PRECISION = 2


SCHEMAS = {
    'rgb': Schema(
        format=Rgb((0, 255), (0, 255), (0, 255)),
        names=Rgb('red', 'green', 'blue'),
        length=3,
        rng=lambda a, b: randint(a, b),
        input_type=int,
        output_type=Rgb,
        validation_type='values',
    ),
    'hsl': Schema(
        format=Hsl((0, 360.0), (0, 100.0), (0, 100.0)),
        names=Hsl('hue', 'saturation', 'lightness'),
        length=3,
        rng=lambda a, b: round(uniform(a, b), PRECISION),
        input_type=float,
        output_type=Hsl,
        validation_type='values',
    ),
    'hsv': Schema(
        format=Hsv((0, 360.0), (0, 100.0), (0, 100.0)),
        names=Hsv('hue', 'saturation', 'value'),
        length=3,
        rng=lambda a, b: round(uniform(a, b), PRECISION),
        input_type=float,
        output_type=Hsv,
        validation_type='values',
    ),
    'ryb': Schema(
        format=Ryb((0, 255), (0, 255), (0, 255)),
        names=Rgb('red', 'yellow', 'blue'),
        length=3,
        rng=lambda a, b: randint(a, b),
        input_type=int,
        output_type=Ryb,
        validation_type='values',
    ),
    'hex': Schema(
        format=re.compile(r''.join([
            '^#?(?:0x)?(?:',  # optional '#' or '0x'
            '(?P<hex>[0-9a-fA-F]{6})|',  # -> #12bc56
            '(?P<hex_short>[0-9a-fA-F]{3})|',  # -> #1b5
            '(?:(?P<hex_alpha>[0-9a-fA-F]{6})[0-9a-fA-F]{2})',  # -> #12bc56FF
            ')$'
        ])),
        names=tuple(),
        length=0,
        rng='#' + ''.join(choice('0123456789ABCDEF') for _ in range(6)),
        input_type=str,
        output_type=lambda x: f'#{x.upper()}',
        validation_type='string',
    ),
    'name': Schema(
        format=color_names,
        names=tuple(),
        length=0,
        rng=choice(list(color_names.keys())),
        input_type=str,
        output_type=str,
        validation_type='string',
    ),
}



import itertools

from acrylic.Defaults import RANDOM, PRECISION, SCHEMAS


def validate(value, colorspace):
    if SCHEMAS[colorspace].validation_type == 'values':
        return validate_values(value, colorspace)
    if SCHEMAS[colorspace].validation_type == 'string':
        return validate_string(value, colorspace)


def in_range(x, a, b, p):
    if x < a or x > b:
        raise ValueError(f'{p!r} should be in range {a} - {b}')
    return x


def check_datatype(datatype, value, param):
    try:
        return datatype(value)
    except ValueError:
        msg = f'cannot convert value {value!r} given for {param!r} '
        msg += f'to {datatype.__name__}'
        raise ValueError(msg) from None
    except TypeError:
        msg = f'cannot convert object of type {type(value).__name__!r} '
        msg += f'given for {param!r} to {datatype.__name__}'
        raise TypeError(msg) from None


def check_iter(iterable, length, param):
    try:
        values = tuple(itertools.islice(iterable, length + 1))
    except TypeError:
        cls = type(iterable).__name__
        msg = f'{cls!r} object given for {param!r} is not iterable'
        raise TypeError(msg) from None

    try:
        values[length - 1]
    except IndexError:
        msg = f'{param!r} should have {length} values'
        raise ValueError(msg) from None

    try:
        values[length]
    except IndexError:
        return values
    else:
        msg = f'{param!r} should have {length} values'
        raise ValueError(msg) from None


def validate_values(values, colorspace):
    schema = SCHEMAS[colorspace]
    datatype = schema.input_type

    if values == RANDOM:
        return [schema.rng(a, b) for a, b in schema.format]

    values = check_iter(values, schema.length, colorspace)

    validated = list()
    for x, (a, b), p in zip(values, schema.format, schema.names):
        value = None

        if x == RANDOM:
            validated.append(schema.rng(a, b))
            continue

        try:
            value = check_datatype(datatype, x, p)
        except TypeError as err:
            try:
                limits = check_iter(x, 2, p)
            except TypeError:
                raise err
            except ValueError:
                msg = f'{p!r} when given as a range needs 2 values'
                raise ValueError(msg) from None
        else:
            if datatype != float:
                validated.append(in_range(value, a, b, p))
            else:
                validated.append(round(in_range(value, a, b, p), PRECISION))
            continue

        limits = [check_datatype(datatype, x, 'r') for x in limits]

        if limits == [-1, -1]:
            validated.append(schema.rng(a, b))
            continue

        limits = [in_range(x, a, b, p) if x != RANDOM else x for x in limits]

        if limits[0] == RANDOM:
            limits[0] = schema.rng(a, limits[1])
        if limits[1] == RANDOM:
            limits[1] = schema.rng(limits[0], b)

        validated.append(schema.rng(min(limits), max(limits)))

    return schema.output_type(*validated)


def validate_string(value, colorspace):
    schema = SCHEMAS[colorspace]

    if value == RANDOM:
        return schema.rng

    value = check_datatype(schema.input_type, value, colorspace)

    if colorspace == 'hex':
        match = schema.format.match(value, 0, 10)
        if match and match.lastgroup == 'hex':
            return schema.output_type(match.group('hex'))
        elif match and match.lastgroup == 'hex_alpha':
            return schema.output_type(match.group('hex_alpha'))
        elif match and match.lastgroup == 'hex_short':
            extended = ''.join(f'{x}{x}' for x in match.group('hex_short'))
            return schema.output_type(extended)

    if colorspace == 'name':
        value = ''.join(value.strip().lower().split())
        if value in schema.format:
            return schema.output_type(value)

    raise ValueError(f'{value!r} is not a valid value for {colorspace!r}')
